/*
 * gigabyte_wmi.c - Gigabyte WMI Hotkey Driver
 *
 * Copyright (C) 2017 Kai-Heng Feng <kai.heng.feng@canonical.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#define pr_fmt(fmt) KBUILD_MODNAME ": " fmt

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/acpi.h>
#include <linux/input.h>
#include <linux/input/sparse-keymap.h>
#include <linux/dmi.h>

MODULE_AUTHOR("Kai-Heng Feng");
MODULE_DESCRIPTION("Gigabyte WMI Hotkey Driver");
MODULE_LICENSE("GPL");

#define WMI_AMW0_GUID0 "ABBC0F6C-8EA1-11D1-00A0-C90629100000"
#define WMI_AMW0_GUID1 "ABBC0F6F-8EA1-11D1-00A0-C90629100000"
#define WMI_AMW0_GUID2 "ABBC0F75-8EA1-11D1-00A0-C90629100000"
#define WMI_EVENT_GUID "ABBC0F72-8EA1-11D1-00A0-C90629100000"

MODULE_ALIAS("wmi:"WMI_AMW0_GUID0);

struct event_return_value {
	u8 function;
	u8 key_num;
	u16 device_state;
	u32 reserved;
} __attribute__((packed));

static struct input_dev *gigabyte_wmi_input_dev;

static const struct key_entry gigabyte_wmi_keymap[] __initconst = {
	/* TODO: Add keymap values once found... */
	/*{ KE_KEY, 0x00, { KEY_ } },*/
	{ KE_END, 0 }
};

static void gigabyte_wmi_notify(u32 value, void *context)
{
	struct acpi_buffer response = { ACPI_ALLOCATE_BUFFER, NULL };
	union acpi_object *obj;
	acpi_status status;
	struct event_return_value return_value;
	u16 device_state;
	const struct key_entry *key;
	u32 scancode;

	status = wmi_get_event_data(value, &response);
	if (ACPI_FAILURE(status)) {
		pr_err("Bad event status 0x%x\n", status);
		return;
	}

	obj = (union acpi_object *)response.pointer;
	if (!obj)
		return;
	if (obj->type != ACPI_TYPE_BUFFER) {
		pr_warn("Unknown response received %d\n", obj->type);
		kfree(obj);
		return;
	}
	if (obj->buffer.length != 8) {
		pr_warn("Unknown buffer length %d\n", obj->buffer.length);
		kfree(obj);
		return;
	}

	return_value = *((struct event_return_value *)obj->buffer.pointer);
	kfree(obj);

	pr_warn("function: 0x%x, key: 0x%x\n", return_value.function,
			return_value.key_num);
}

static struct dmi_system_id gigabyte_wmi_dmi_table[] __initdata = {
	{
		.ident = "Gigabyte P57",
		.matches = {
			DMI_MATCH(DMI_SYS_VENDOR, "GIGABYTE"),
			DMI_MATCH(DMI_PRODUCT_NAME, "P57"),
		},
	},
	{}
};

static int __init gigabyte_wmi_input_setup(void)
{
	acpi_status status;
	int err;

	gigabyte_wmi_input_dev = input_allocate_device();
	if (!gigabyte_wmi_input_dev)
		return -ENOMEM;

	gigabyte_wmi_input_dev->name = "Gigabyte WMI hotkeys";
	gigabyte_wmi_input_dev->phys = "wmi/input0";
	gigabyte_wmi_input_dev->id.bustype = BUS_HOST;

	err = sparse_keymap_setup(gigabyte_wmi_input_dev,
				  gigabyte_wmi_keymap, NULL);
	if (err)
		goto err_free_dev;

	status = wmi_install_notify_handler(WMI_EVENT_GUID,
					    gigabyte_wmi_notify, NULL);
	if (ACPI_FAILURE(status)) {
		err = -EIO;
		goto err_free_keymap;
	}

	err = input_register_device(gigabyte_wmi_input_dev);
	if (err)
		goto err_remove_notifier;

	return 0;

 err_remove_notifier:
	wmi_remove_notify_handler(WMI_EVENT_GUID);
 err_free_keymap:
	sparse_keymap_free(gigabyte_wmi_input_dev);
 err_free_dev:
	input_free_device(gigabyte_wmi_input_dev);
	return err;
}

static void gigabyte_wmi_input_destroy(void)
{
	wmi_remove_notify_handler(WMI_EVENT_GUID);
	sparse_keymap_free(gigabyte_wmi_input_dev);
	input_unregister_device(gigabyte_wmi_input_dev);
}

static int __init gigabyte_wmi_init(void)
{
	int ret;

	if (!wmi_has_guid(WMI_AMW0_GUID0) ||
	    !dmi_check_system(gigabyte_wmi_dmi_table)) {
		pr_err("Failed to match DMI or WMI\n");
		return -ENODEV;
	}

	ret = gigabyte_wmi_input_setup();
	if (ret) {
		pr_err("Failed to setup input device\n");
		return ret;
	}

	pr_warn("Gigabyte WMI Hotkey Driver\n");

	return 0;
}

static void __exit gigabyte_wmi_exit(void)
{
	if (wmi_has_guid(WMI_EVENT_GUID))
		gigabyte_wmi_input_destroy();
}

module_init(gigabyte_wmi_init);
module_exit(gigabyte_wmi_exit);
